
void resetIndentInfo(){
  indentLevel = 0;
  indentLevelExec = 0;
  for(int i=0; i<MAX_INDENTS; i++){
    indentWasTrue[i] = true;
    jumpLine[i] = -1;
  }
  requestedJumpLine = -1;
}

void runPage(int p, int initialLineOffsetOnPage, boolean justJumped){
  int offset = EEPROM_PROG_OFFSET + 2*p*WORDS_PER_PAGE;
  int prevLine = -1;
  if(justJumped){
    // if we just jumped, don't count the first line as and reset linesSinceLastRead.
    linesSinceLastRead = 0;
    prevLine = initialLineOffsetOnPage;
  }
  currentPage = p;
  //Serial.print("RUN PAGE >> pg_");
  //Serial.print(currentPage);
  //Serial.print(" offset=");
  //Serial.println(initialLineOffsetOnPage);
  // now, Evaluate the Code Page!
  for(int i=initialLineOffsetOnPage*4; (i+3)<WORDS_PER_PAGE; i+=4){
    currentLineOnPage = i/4;
    while(!isGoing && !deviceHasBeenReset){
      // PAUSE EXECUTION UNTIL WE'RE GOING.
      serialSafeDelay(20);
    }
    if(deviceHasBeenReset){
      break;
    }
    if(prevLine != currentLineOnPage){
      linesSinceLastRead++;
    }
    prevLine = currentLineOnPage;
    int ins1 = EEPROM.read(offset + i*2);
    int ins2 = EEPROM.read(offset + i*2 + 1);
    int ins = ins1<<8 | ins2;
    int insA = ((EEPROM.read(offset + (i+1)*2)<<8) & 0xFF00) | (EEPROM.read(offset + (i+1)*2+1) & 0x00FF);
    int insB = ((EEPROM.read(offset + (i+2)*2)<<8) & 0xFF00) | (EEPROM.read(offset + (i+2)*2+1) & 0x00FF);
    int insC = ((EEPROM.read(offset + (i+3)*2)<<8) & 0xFF00) | (EEPROM.read(offset + (i+3)*2+1) & 0x00FF);
    switch(ins){
    case INS_HEX_NOOP:
      {
        // No Op.  don't do anything.
        break;
      }  
    case INS_HEX_OUT | 0x0000:
    case INS_HEX_OUT | 0x0001:
    case INS_HEX_OUT | 0x0002:
    case INS_HEX_OUT | 0x0003:
    case INS_HEX_OUT | 0x0004:
    /*
    case INS_HEX_OUT | 0x0005:
    case INS_HEX_OUT | 0x0006:
    case INS_HEX_OUT | 0x0007:
    */
      {
        if(indentWasTrue[indentLevelExec]){
          //// Serial.println("Setting output.");
          setOutput(ins2 & 0x0F, evaluateVal(insA), insB);
        }
        break;
      }
    case INS_HEX_DEF | 0x0000:
    case INS_HEX_DEF | 0x0001:
    case INS_HEX_DEF | 0x0002:
    case INS_HEX_DEF | 0x0003:
    case INS_HEX_DEF | 0x0004:
    case INS_HEX_DEF | 0x0005:
    case INS_HEX_DEF | 0x0006:
    case INS_HEX_DEF | 0x0007:
    case INS_HEX_DEF | 0x0008:
      {
        if(indentWasTrue[indentLevelExec]){
          //// Serial.println("Defining variable.");
          setTempVar(ins2 & 0x0F, evaluateVal(insA), insB, evaluateVal(insC));
        }
        break;
      }
    case INS_HEX_DELAY:
      {
        // Delay. (with safe serial, so we never lose control)
        if(indentWasTrue[indentLevelExec]){
          //// Serial.println("Delay.");
          serialSafeDelay(evaluateVal(insA));
        }
        break;
      }
    case INS_HEX_IF:
      {
        // start an if statement.
        if(indentWasTrue[indentLevelExec]){
          //// Serial.println("If.");
          indentLevelExec += 1;
          if( evalConditional(evaluateVal(insA), insB, evaluateVal(insC)) ){
            indentWasTrue[indentLevelExec] = true;
          }
          else{
            indentWasTrue[indentLevelExec] = false;
          }
        }
        indentLevel += 1;
        break;
      }
    case INS_HEX_ELSE:
      {
        // use an else statement.
        if(indentLevelExec == indentLevel){
          //// Serial.println("Else.");
          if(indentWasTrue[indentLevelExec]){
            indentWasTrue[indentLevelExec] = false;
          }
          else{
            indentWasTrue[indentLevelExec] = true;
          }
        }
        break;
      }
    case INS_HEX_ENDIF:
      {
        // end an if statement.
        if(indentLevelExec == indentLevel){
          //Serial.print("End if. level=");
          //Serial.println(indentLevelExec);
          if(jumpLine[indentLevelExec] >= 0){
            // this end is actually closing a while loop.  jump back.
            //Serial.print("JMP >> ENDIF");
            requestedJumpLine = jumpLine[indentLevelExec];
            jumpLine[indentLevelExec] = -1; // reset jumpline.
            i = WORDS_PER_PAGE; // skip anything else on this page, we're in the process of jumping.
          }
          indentLevelExec -= 1;
        }
        indentLevel -= 1;
        break;
      }
    case INS_HEX_WHILE:
      {
        if(indentWasTrue[indentLevelExec]){
          indentLevelExec += 1;
          if( evalConditional(evaluateVal(insA), insB, evaluateVal(insC)) ){
            // continue normally following the indent...
            indentWasTrue[indentLevelExec] = true;
            // save page and offset for jumping back to this instruction at the indentation end.
            jumpLine[indentLevelExec] = currentPage*(LINES_PER_PAGE) + currentLineOnPage;
            //Serial.print("JMP >> set in while >> level=");
            //Serial.print(indentLevelExec);
            //Serial.print(" jumpline=");
            //Serial.println(jumpLine[indentLevelExec]);
          }
          else{
            // don't execute the indent
            indentWasTrue[indentLevelExec] = false;
          }
        }
        indentLevel += 1;
        break;
      }
    case INS_HEX_WAIT:
      {
        if(indentWasTrue[indentLevelExec]){
          if( evalConditional(evaluateVal(insA), insB, evaluateVal(insC)) ){
            // continue normally...
          }
          else{
            // stay at this instruction... (-4 so that we'll repeat it).
            serialSafeDelay(10);
            i -= 4;
          }
        }
        break;
      }
    default: 
      {
        Serial.print("ERR >> RUN -- ins? >> 0x");
        Serial.println(ins, HEX);
      }
    } // close switch
  }
}

void setTempVar(int t, int A, int op, int C){
  if(t >= 0 && t < TOTAL_TEMPVARS){
    switch(op){
    case 0xE000:
    case 0xE005:
      { // addition.
        tempVals[t] = A + C;
        break;
      }
    case 0xE001:
      { // subtraction.
        tempVals[t] = A - C;
        break;
      }
    case 0xE002:
      { // multiply.
        tempVals[t] = A * C;
        break;
      }
    case 0xE003:
      { // divide.
        if(C == 0){
          Serial.println("WARN >> Div(/) by Zero.");
        }
        tempVals[t] = A / C;
        break;
      }
    case 0xE004:
      { // mod.
        if(C == 0){
          Serial.println("WARN >> Mod(%) by Zero.");
        }
        tempVals[t] = A % C;
        break;
      }
    default:
      {
        // Serial.print("SetTemp >> Error, op 0x");
        // Serial.println(op, HEX);
      }
    }
  }
  else{
    // Serial.print("SetTemp >> Error, t ");
    // Serial.println(t);
  }
}


boolean evalConditional(int a, int comp, int c){
  if(comp == 0xE000){
    return (a == c);
  }
  if(comp == 0xE001){
    return (a <  c);
  }
  if(comp == 0xE002){
    return (a <= c);
  }
  if(comp == 0xE003){
    return (a >  c);
  }
  if(comp == 0xE004){
    return (a >= c);
  }
  if(comp == 0xE005){
    return (a != c);
  }
  return true; 
}

int evaluateVal(int ins){
  int h = 0;
  // return a numerical value for the variable.
  if((ins & 0xF000) == 0x9000){
    // Random
    h = random(0, 1001); // 0 - 1000 inclusive
    return h;  
  }
  if((ins & 0xFF00) == 0xA000){
    // Input.
    updateInputs();
    h = ins&0x007F;
    if(h >= 0 && h < TOTAL_INPUTS){
      return inputVals[h];
    }
    else{
      // Serial.print("EVAL >> Error, Input ");
      // Serial.println(h);
      return 0;
    }   
  }
  if((ins & 0xFF00) == 0xB000){
    // Output.
    h = ins&0x007F;
    if(h >= 0 && h < TOTAL_OUTPUTS){
      return outputVals[h];
    }
    else{
      // Serial.print("EVAL >> Error, Output ");
      // Serial.println(h);
      return 0;
    }   
  }
  if((ins & 0xFF00) == 0xC000){
    // Temp VAR
    h = ins&0x007F;
    if(h >= 0 && h < TOTAL_TEMPVARS){
      return tempVals[h];
    }
    else{
      // Serial.print("EVAL >> Error, Temp ");
      // Serial.println(h);
      return 0;
    }   
  }
  if((ins & 0xF000) == 0xD000){
    // CONST
    h = ins&0x007F;
    h = h | ((ins & 0x0F00)>>1);
    h -= 1024;
    return h;  
  }
  // Serial.println("EVAL >> Error.");
  return 0;
}


void setOutput(int port, int val, int type){
  if(port >= 0 && port < TOTAL_OUTPUTS){
    // TODO: actually set the value, and make sure it's the right output type.
    outputVals[port] = min(1000, max(0, val));
    if(outputType[port] != type){
      setOutputType(port, type);
    }
    //updateOutputs();
    // Now UPDATE THE OUTPUT!
    if(outputType[port] == 0xE000){
      // TYPE = DIG
      if(outputVals[port] > 500){
        setOutputFreqPeriodAndOnFor(port, 0, 100, 100);
      }
      else{
        setOutputFreqPeriodAndOnFor(port, 0, 100, 0);
      }
    }
    if(outputType[port] == 0xE001 || outputType[port] == 0xE004){
      // TYPE = PWM
      if(outputVals[port] < 0){
        setOutputFreqPeriodAndOnFor(port, 0, 255, 0);
      }
      else{
        if(outputVals[port] > 1000){
          setOutputFreqPeriodAndOnFor(port, 0, 255, 255);
        }
        else{
          float f = outputVals[port]/3.92;
          setOutputFreqPeriodAndOnFor(port, 0, 255, max(0, min(255, ((int)f))) );
        }
      }
    }
    if(outputType[port] == 0xE002 || outputType[port] == 0xE005){
      // TYPE = TONE
      if(outputVals[port] < 1){
        setOutputFreqPeriodAndOnFor(port, 1, 2, 1);
      }
      else{
        if(outputVals[port] > 10000){
          setOutputFreqPeriodAndOnFor(port, 10000, 2, 1);
        }
        else{
          setOutputFreqPeriodAndOnFor(port, outputVals[port], 2, 1);
        }
      }
    }
    if(outputType[port] == 0xE003){
      // TYPE = SRV
      if(outputVals[port] <= 0){
        setOutputFreqPeriodAndOnFor(port, 0, 800, 24);
      }
      else{
        if(outputVals[port] >= 1000){
          setOutputFreqPeriodAndOnFor(port, 0, 800, 90);
        }
        else{
          setOutputFreqPeriodAndOnFor(port, 0, 800, 24 + outputVals[port]/15);
        }
      }
    }

  }
}

void updateInputs(){
  inputVals[0] = max(0, min(1000, analogRead(PIN_IN_0)-18));
  inputVals[1] = max(0, min(1000, analogRead(PIN_IN_1)-18));
  inputVals[2] = max(0, min(1000, analogRead(PIN_IN_2)-18));
  inputVals[3] = max(0, min(1000, analogRead(PIN_IN_3)-18));
  inputVals[4] = max(0, min(1000, analogRead(PIN_IN_4)-18));
  //inputVals[5] = max(0, min(1000, analogRead(PIN_IN_5)-12));
  //inputVals[6] = max(0, min(1000, analogRead(PIN_IN_6)-12));
  //inputVals[7] = max(0, min(1000, analogRead(PIN_IN_7)-12));
}

void setOutputType(int port, int type){
  outputType[port] = type;
}

void setOutputFreqPeriodAndOnFor(int port, int f, int period, int onFor){
  float t;
  unsigned int fBound = min(20000, max(1, f));
  int set_onFor  = 0;
  int set_period = 2;
  if(f == 0){
    set_onFor  = onFor;
    set_period = period; 
  }
  else{
    t = 4000.0/fBound;
    set_onFor  = t*5;
    set_period = set_onFor*2;
  }
  switch(port){
  case 0:
    {
      outputOnFor_0  = set_onFor;
      outputPeriod_0 = set_period;
      break;
    }
  case 1:
    {
      outputOnFor_1  = set_onFor;
      outputPeriod_1 = set_period;
      break;
    }
  case 2:
    {
      outputOnFor_2  = set_onFor;
      outputPeriod_2 = set_period;
      break;
    }
  case 3:
    {
      outputOnFor_3  = set_onFor;
      outputPeriod_3 = set_period;
      break;
    }
  case 4:
    {
      outputOnFor_4  = set_onFor;
      outputPeriod_4 = set_period;
      break;
    }
  case 5:
    {
      outputOnFor_5  = set_onFor;
      outputPeriod_5 = set_period;
      break;
    }
  case 6:
    {
      outputOnFor_6  = set_onFor;
      outputPeriod_6 = set_period;
      break;
    }
  case 7:
    {
      outputOnFor_7  = set_onFor;
      outputPeriod_7 = set_period;
      break;
    }
  }
}



