
//  Teagueduino
//  Copyright (C) 2011  Teague Labs / Adam Kumpf (akumpf@teague.com)
//
//  http://labs.teague.com, http://teagueduino.org
//
//  This program is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.

//  This program is distributed in the hope that it will be useful, 
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//


#define PROTO_LENGTH_MAX  64
#define PROTO_LENGTH_DATA 512
#define PROTO_CHAR_START ':'
#define PROTO_CHAR_END   ';'

#define PROTO_CMD_INFO_GET  'i'
#define PROTO_CMD_INFO_SET  'I'
#define PROTO_CMD_NOTES_GET 'n'
#define PROTO_CMD_NOTES_SET 'N'
#define PROTO_CMD_DATA_GET  'd'
#define PROTO_CMD_DATA_SET  'D'
#define PROTO_CMD_PAGE_GET  'p'
#define PROTO_CMD_PAGE_SET  'P'
#define PROTO_CMD_RESET     'R'
#define PROTO_CMD_WORD_SET  'W'
#define PROTO_CMD_WORD_SET_CONFIRM 'w' // confirmation
#define PROTO_CMD_GOING_SET 'G'

#define PIN_IN_0 7
#define PIN_IN_1 6
#define PIN_IN_2 5
#define PIN_IN_3 4
#define PIN_IN_4 3
#define PIN_IN_5 2
#define PIN_IN_6 1
#define PIN_IN_7 0

#define PIN_INLED_0 19
#define PIN_INLED_1 20
#define PIN_INLED_2 21
#define PIN_INLED_3 22
#define PIN_INLED_4 23
#define PIN_INLED_5 24
#define PIN_INLED_6 25
#define PIN_INLED_7 26

#define PIN_OUT_0 17
#define PIN_OUT_1 16
#define PIN_OUT_2 15
#define PIN_OUT_3 14
#define PIN_OUT_4 13
#define PIN_OUT_5 12
#define PIN_OUT_6 11
#define PIN_OUT_7 10

#define INS_HEX_NOOP  0x8000 //
#define INS_HEX_IF    0x8010 // if(A,B,C){
#define INS_HEX_ELSE  0x8011 // }else{
#define INS_HEX_ENDIF 0x8012 // }
#define INS_HEX_OUT   0x8020 // out(A,B,C)
#define INS_HEX_DEF   0x8030 // V1,V2,V3,V4 = A,B,C
#define INS_HEX_DELAY 0x8040 // delay(A)
#define INS_HEX_FOR   0x8050 // for(A,B)
#define INS_HEX_WAIT  0x8060 // waituntil(A,B,C)
#define INS_HEX_WHILE 0x8070 // while(A,B,C)

//HardwareSerial Uart = HardwareSerial();

// helpful for clearing and setting specific bits.
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif 

#define TOTAL_PAGES 13 // pages of code (last page is closing brackets).

#include <EEPROM.h>

#define EEPROM_PAGE_SIZE         48*2
#define EEPROM_PROG_OFFSET       0
//#define EEPROM_NAME_OFFSET       EEPROM_PAGE_SIZE*TOTAL_PAGES
//#define EEPROM_NAME_SIZE         40
#define EEPROM_VERSION_OFFSET    EEPROM_PAGE_SIZE*TOTAL_PAGES
#define EEPROM_NOTES_OFFSET      EEPROM_PAGE_SIZE*(TOTAL_PAGES+1)

#define EEPROM_NOTES_PAGES_NAMEAUTH  1
#define EEPROM_NOTES_PAGES_COMMENTS  6
#define EEPROM_NOTES_PAGES_IO_NOTES  10
#define EEPROM_NOTES_PAGES           EEPROM_NOTES_PAGES_NAMEAUTH + EEPROM_NOTES_PAGES_COMMENTS + EEPROM_NOTES_PAGES_IO_NOTES


#define WORDS_PER_SECTION        12
#define SECTIONS_PER_PAGE        4
#define WORDS_PER_PAGE           WORDS_PER_SECTION*SECTIONS_PER_PAGE
#define LINES_PER_PAGE           WORDS_PER_PAGE/4


#define VERSION_NUM              3 // change this if you want the EEPROM to be reflashed.


unsigned int eepromVal = 0;

#include <pins_arduino.h>

#define TOTAL_INPUTS   5
#define TOTAL_OUTPUTS  5
#define TOTAL_TEMPVARS 9

volatile int inputVals [TOTAL_INPUTS] = {
  0, 0, 0, 0, 0 };
int outputVals[TOTAL_OUTPUTS] = {
  0, 0, 0, 0, 0 };
int outputType[TOTAL_OUTPUTS] = {
  0, 0, 0, 0, 0 };
int tempVals  [TOTAL_TEMPVARS] = {
  0, 0, 0, 0, 0, 0, 0, 0, 0 };

/*
volatile unsigned int outputPeriod[8]  = {
 2, 2, 2, 2, 2, 2, 2, 2 };
 */
int outputPeriod_0 = 2;
int outputPeriod_1 = 2;
int outputPeriod_2 = 2;
int outputPeriod_3 = 2;
int outputPeriod_4 = 2;
int outputPeriod_5 = 2;
int outputPeriod_6 = 2;
int outputPeriod_7 = 2;
/*
volatile unsigned int outputOnFor[8]   = {
 0, 0, 0, 0, 0, 0, 0, 0 };
 */
int outputOnFor_0 = 0;
int outputOnFor_1 = 0;
int outputOnFor_2 = 0;
int outputOnFor_3 = 0;
int outputOnFor_4 = 0;
int outputOnFor_5 = 0;
int outputOnFor_6 = 0;
int outputOnFor_7 = 0;
/*
volatile unsigned int outputCounter[8] = {
 0, 0, 0, 0, 0, 0, 0, 0 }; 
 */
int outputCounter_0 = 0;
int outputCounter_1 = 0;
int outputCounter_2 = 0;
int outputCounter_3 = 0;
int outputCounter_4 = 0;
int outputCounter_5 = 0;
int outputCounter_6 = 0;
int outputCounter_7 = 0;

int inputLEDCounter = 1;

int inputLEDAlternator = 0;

// these variables are for parsing/running the code! :)
int indentLevelExec = 0; // the actual level the code is executing.
int indentLevel = 0;     // the level the code is at (may be passed over).
#define MAX_INDENTS 12
boolean indentWasTrue[MAX_INDENTS];
int     jumpLine[MAX_INDENTS];
int     requestedJumpLine = -1;

int OUT_PINS[4] = { 
  PIN_OUT_0, PIN_OUT_1, PIN_OUT_2, PIN_OUT_3};
  
#define setOut0_HIGH() sbi(PORTC,7)
#define setOut0_LOW()  cbi(PORTC,7)
#define setOut1_HIGH() sbi(PORTC,6)
#define setOut1_LOW()  cbi(PORTC,6)
#define setOut2_HIGH() sbi(PORTC,5)
#define setOut2_LOW()  cbi(PORTC,5)
#define setOut3_HIGH() sbi(PORTC,4)
#define setOut3_LOW()  cbi(PORTC,4)
#define setOut4_HIGH() sbi(PORTC,3)
#define setOut4_LOW()  cbi(PORTC,3)
#define setOut5_HIGH() sbi(PORTC,2)
#define setOut5_LOW()  cbi(PORTC,2)
#define setOut6_HIGH() sbi(PORTC,1)
#define setOut6_LOW()  cbi(PORTC,1)
#define setOut7_HIGH() sbi(PORTC,0)
#define setOut7_LOW()  cbi(PORTC,0)

#define setInLED0_HIGH() sbi(PORTE, 7)
#define setInLED0_LOW()  cbi(PORTE, 7)
#define setInLED1_HIGH() sbi(PORTB, 0)
#define setInLED1_LOW()  cbi(PORTB, 0)
#define setInLED2_HIGH() sbi(PORTB, 1)
#define setInLED2_LOW()  cbi(PORTB, 1)
#define setInLED3_HIGH() sbi(PORTB, 2)
#define setInLED3_LOW()  cbi(PORTB, 2)
#define setInLED4_HIGH() sbi(PORTB, 3)
#define setInLED4_LOW()  cbi(PORTB, 3)
#define setInLED5_HIGH() sbi(PORTB, 4)
#define setInLED5_LOW()  cbi(PORTB, 4)
#define setInLED6_HIGH() sbi(PORTB, 5)
#define setInLED6_LOW()  cbi(PORTB, 5)
#define setInLED7_HIGH() sbi(PORTB, 6)
#define setInLED7_LOW()  cbi(PORTB, 6)


#define setDebugLightOn()  sbi(PORTD, 6)
#define setDebugLightOff() cbi(PORTD, 6)


// for handing incoming serial data...
byte protocolIndex     = 0;
byte protocolCmd       = 0;
byte protocolLength    = 0;
byte protocolDataIndex = 0;
byte protocolData[PROTO_LENGTH_DATA];

int currentPage = 0;
int currentLineOnPage = 0;
int linesSinceLastRead = 0;

boolean deviceHasBeenReset = true;

unsigned int preOCR = 0;

//boolean running = true; // used internally for reset.

boolean isGoing = true; // user controller, for pausing execution.

void setup(){
  Serial.begin(115200);
  randomSeed(analogRead(0)+analogRead(1)+analogRead(2)+analogRead(3));
}
//boolean okayToRunMainLoop = false;
void setup_Step2(){
  TIMSK1 = 0x00; // disable compare interrupts A, B, and C.
  
  protocolIndex     = 0;
  protocolCmd       = 0;
  protocolLength    = 0;
  protocolDataIndex = 0;
  indentLevelExec = 0; // the actual level the code is executing.
  indentLevel = 0;
  currentPage = 0;
  currentLineOnPage = 0;
  linesSinceLastRead = 0;
  for(int i=0; i<TOTAL_INPUTS; i++){
    inputVals[i]  = 0;
    outputVals[i] = 0;
    outputType[i] = 0;
  }
  for(int i=0; i<TOTAL_TEMPVARS; i++){
    tempVals[i]   = 0;
  }
  outputPeriod_0 = 2;
  outputPeriod_1 = 2;
  outputPeriod_2 = 2;
  outputPeriod_3 = 2;
  outputPeriod_4 = 2;
  outputPeriod_5 = 2;
  outputPeriod_6 = 2;
  outputPeriod_7 = 2;
  outputOnFor_0 = 0;
  outputOnFor_1 = 0;
  outputOnFor_2 = 0;
  outputOnFor_3 = 0;
  outputOnFor_4 = 0;
  outputOnFor_5 = 0;
  outputOnFor_6 = 0;
  outputOnFor_7 = 0;
  outputCounter_0 = 0;
  outputCounter_1 = 0;
  outputCounter_2 = 0;
  outputCounter_3 = 0;
  outputCounter_4 = 0;
  outputCounter_5 = 0;
  outputCounter_6 = 0;
  outputCounter_7 = 0;
  // enable internal pullups on Analog inputs.
  /*
  digitalWrite(38 + PIN_IN_0, HIGH);
  digitalWrite(38 + PIN_IN_1, HIGH);
  digitalWrite(38 + PIN_IN_2, HIGH);
  digitalWrite(38 + PIN_IN_3, HIGH);
  digitalWrite(38 + PIN_IN_4, HIGH);
  digitalWrite(38 + PIN_IN_5, HIGH);
  digitalWrite(38 + PIN_IN_6, HIGH);
  diitalWrite(38 + PIN_IN_7, HIGH);
  */
  reflashEEPROMIfNeeded();
  Serial.flush();

  pinMode(PIN_OUT_0, OUTPUT);
  pinMode(PIN_OUT_1, OUTPUT);
  pinMode(PIN_OUT_2, OUTPUT);
  pinMode(PIN_OUT_3, OUTPUT);
  pinMode(PIN_OUT_4, OUTPUT);
  pinMode(PIN_OUT_5, OUTPUT);
  pinMode(PIN_OUT_6, OUTPUT);
  pinMode(PIN_OUT_7, OUTPUT);

  delay(500);

  digitalWrite(PIN_OUT_0, LOW);
  digitalWrite(PIN_OUT_1, LOW);
  digitalWrite(PIN_OUT_2, LOW);
  digitalWrite(PIN_OUT_3, LOW);
  digitalWrite(PIN_OUT_4, LOW);
  digitalWrite(PIN_OUT_5, LOW);
  digitalWrite(PIN_OUT_6, LOW);
  digitalWrite(PIN_OUT_7, LOW);

  pinMode(PIN_INLED_0, OUTPUT);
  pinMode(PIN_INLED_1, OUTPUT);
  pinMode(PIN_INLED_2, OUTPUT);
  pinMode(PIN_INLED_3, OUTPUT);
  pinMode(PIN_INLED_4, OUTPUT);
  pinMode(PIN_INLED_5, OUTPUT);
  pinMode(PIN_INLED_6, OUTPUT);
  pinMode(PIN_INLED_7, OUTPUT);
  digitalWrite(PIN_INLED_0, LOW);
  digitalWrite(PIN_INLED_1, LOW);
  digitalWrite(PIN_INLED_2, LOW);
  digitalWrite(PIN_INLED_3, LOW);
  digitalWrite(PIN_INLED_4, LOW);
  digitalWrite(PIN_INLED_5, LOW);
  digitalWrite(PIN_INLED_6, LOW);
  digitalWrite(PIN_INLED_7, LOW);

  // debug light.
  pinMode(6, OUTPUT);

  showStartupLights();

  // OUT 0 on Timer1
  TCCR1A = 0x00;   // Normal mode (no PWM).
  TCCR1B = 0x0A;   // CTC (compare timer, then clear) , set the Timer1 clock divide by 2.
  OCR1AH = 0x00;   // CTC H  -- 0x00C5 = 10kHz (5kHz tone), 0x0020 = 60kHz (30kHz tone
  OCR1AL = 0x31;   // CTC L
  TIMSK1 = 0x02; // enable compare interrupts A, B, and C.
}

void loop(){
  // SETUP IF NEEDED
  int pageLineOffset = 0;
  boolean justJumped = false;
  if(deviceHasBeenReset){
    isGoing = true;
    deviceHasBeenReset = false;
    setup_Step2(); // initialize device.
    resetIndentInfo();
    for(int i=0; i<4; i++){
      handleSerialDataIn();
      runPage(i, pageLineOffset, justJumped);  // run the user-specified setup code.
      pageLineOffset = 0; // start the next page fresh from the start.
      justJumped = false; // back to normal
      if(requestedJumpLine >= 0){
        // we're jumping!
        int page = min(3, max(0, requestedJumpLine/LINES_PER_PAGE));
        pageLineOffset = min(LINES_PER_PAGE-1, max(0, requestedJumpLine%LINES_PER_PAGE));
        requestedJumpLine = -1; // done with this jump.
        justJumped = true;
        i = page-1; // i will increment at the end of this for loop...
      }
      updateInputs(); // mainly so user can see green input lights change...
    }
  }
  // MAIN PROGRAM
  resetIndentInfo();
  pageLineOffset = 0;
  justJumped = false;
  for(int i=4; i<TOTAL_PAGES; i++){
    if(deviceHasBeenReset == false){
      handleSerialDataIn();
      runPage(i, pageLineOffset, justJumped);
      pageLineOffset = 0; // start the next page fresh from the start.
      justJumped = false; // back to normal
      if(requestedJumpLine >= 0){
        // we're jumping!
        int page = min(TOTAL_PAGES, max(4, requestedJumpLine/LINES_PER_PAGE));
        pageLineOffset = min(LINES_PER_PAGE-1, max(0, requestedJumpLine%LINES_PER_PAGE));
        requestedJumpLine = -1; // done with this jump.
        linesSinceLastRead = 0;
        justJumped = true;
        i = page-1; // i will increment at the end of this for loop...
      }
      updateInputs(); // mainly so user can see green input lights change...
    }
  }
}





